import { Component, OnInit } from '@angular/core';
import { ActivatedRoute, Params } from '@angular/router';
import { DomSanitizer } from '@angular/platform-browser'

import { Router } from '@angular/router';

import { ViewChild } from '@angular/core';
import { MatMenuTrigger } from '@angular/material'
declare var Viz: any;

@Component({
  selector: 'app-turing-machine',
  templateUrl: './turing-machine.component.html',
  styleUrls: ['./turing-machine.component.css', '../../assets/css/turing.css']
})
export class TuringMachineComponent implements OnInit {
  @ViewChild(MatMenuTrigger) trigger: MatMenuTrigger;

  openMenu() {
    this.trigger.openMenu();
  }

  constructor(
    private activatedRoute: ActivatedRoute,  sanitizer: DomSanitizer,
    private router: Router) {

  }

  // Dynamic parameters for this component's route: /example-params/:first/:second
  routeParams: Params;
  // Query parameters found in the URL: /example-params/one/two?query1=one&query2=two
  queryParams: Params;
  exp_id;
  exp = ["1", "2", "3"]
  simulate_interval = 500;
  one_step_mode = false;
  stop_check = false;
  is_simulating = false;
  speed = 1;
  is_initialized;
  setting;
  self_turing_folder_url;
  test_rules = ""
  test_raw_rules = ""
  editorOptions = { theme: 'vs-dark', language: "text/plain", automaticLayout: true, }
  pauseContinue = "Continue"
  subIDs = ["1"]
  hideQuestions = true
 
  @ViewChild('chart') componentRef;
  chartUpdateFlag;
  async ngOnInit() {
    this.routeParams = {}
    const fetcherOptions = {};
    this.activatedRoute.params.subscribe(params => {
      this.routeParams = params;
      console.log("this.routeParams", this.routeParams)
      this.init();
    });
    // this.init();
  }

  async init() {
    if ("id" in this.routeParams) {

      this.exp_id = this.routeParams["id"];
      if (this.exp_id == "0") { //here is index page
        this.hideQuestions = true
        return;
      }

      if (!this.exp.includes(this.exp_id)) {
        await this.router.navigateByUrl('');
      }
    } else {
      await this.router.navigateByUrl('');
    }
    this.hideQuestions = false
    
    var tape = new Turing_Tape("");
    draw_Tape(tape);
  }
  async QuestionSubIDSelected() {
  
    var tape = new Turing_Tape("");
    draw_Tape(tape);
   
      (document.getElementById("rule_upload_Button") as HTMLButtonElement).disabled = false;
      (document.getElementById("rule_download_Button") as HTMLButtonElement).disabled = false;
    
  }

  init_set() {
    this.is_initialized = true;
    this.setting = new set_env(this.test_raw_rules);
    this.pauseContinue = "Continue"
    print_current_step(this.setting.step);
    // console.log(setting.test_data);
  }
  continue_simulate() {
    if (this.setting.is_pause) {

      return;
    } else {
      // console.log("setting.is_pause",setting.is_pause);
      try {
        this.exe_next_step();
      } catch (error) {
        alert(error);
        return;
      }
      draw_Tape(this.setting.tape);

      setTimeout(() => { this.continue_simulate() }, this.simulate_interval);
    }
  }

  exe_next_step() {
    if (this.setting.current_state == "qa" || this.setting.step > 10000) {
      if (this.setting.step > 10000) {
        throw "The total number of steps exceeds 10000, the turing machine may not stop";
      }
      return 0;
    }
    var data_in = this.setting.tape.currentCharacter();
    this.setting.step++;
    var current_rule = this.setting.rules.find_rule(this.setting.current_state, data_in);
    try {
      if (current_rule == null) {
        this.setting.is_pause = true;
        throw "No corresponding rule found for " +
        "state: " +
        this.setting.current_state +
        " input: " +
        data_in;
      }
      // if(current_rule.data_out != 'N'){
      this.setting.tape.write(current_rule.data_out);
      // }
      this.setting.tape.move(current_rule.operation);
      this.setting.current_state = current_rule.next_state;
      print_current_rule(current_rule);
      print_current_step(this.setting.step);
      var next_data_in = this.setting.tape.currentCharacter();
      var next_rule = this.setting.rules.find_rule(this.setting.current_state, next_data_in);
      print_next_rule(next_rule);
    } catch (error) {
      alert(error);
      return null;
    }
  }

  pause_continue() {
    if (!this.is_initialized) {
      alert("Not initialized, please click 'Set'");
      return;
    }

    if (this.setting.is_pause) {
      this.pauseContinue = "Pause"
      this.setting.is_pause = false;
      this.continue_simulate();
    } else {
      this.pauseContinue = "Continue"
      this.setting.is_pause = true;
    }
  }

  one_step() {
    if (!this.is_initialized) {
      alert("Not initialized, please click 'Set'");
      return;
    }

    if (this.setting.is_pause) {
      this.exe_next_step();
      draw_Tape(this.setting.tape);
    } else {
      return;
    }
  }

  get_simulate_answer() {
    var rules = get_Rules(this.test_raw_rules);
    if (rules == null) {
      return;
    }
    this.init_set();
    var test_data = (document.getElementById("test_data") as HTMLInputElement).value;
    var tape;
    tape = new Turing_Tape(test_data);
    var step, current_state, data_in;
    step = 0;
    try {
      current_state = "q0";
      data_in = tape.currentCharacter();

      while (true) {
        if (current_state == "qa" || step > 10000) {
          if (step > 10000) {
            throw "The total number of steps exceeds 10000, the turing machine may not stop";
          }
          break;
        }

        step++;
        // console.log(current_state, data_in);
        var current_rule = rules.find_rule(current_state, data_in);
        // console.info(current_rule);
        // current_rule.print();
        if (current_rule == null) {
          throw "No corresponding rule found for" +
          "state: " +
          current_state +
          " input:" +
          data_in;
        }
        // if (current_rule.data_out != 'N'){
        tape.write(current_rule.data_out);
        // console.info(tape.printedCharacters)
        // }
        tape.move(current_rule.operation);
        current_state = current_rule.next_state;
        data_in = tape.currentCharacter();
      }
      print_current_step(step);
      var answer = "";
      for (let i = 0; i <= 50; i++) {
        answer += tape.characterAtIndex(i);
      }
      let info = "Result: " + answer;
      alert(info);
      // tape.currentCharacter()
      draw_Tape(tape);
      setTimeout(function () {
        draw_Tape(tape);
      }, 600);
      // setTimeout(function(){draw_Tape(tape);}, 100);
      // }
    } catch (error) {
      alert(error);
      return;
    }
  }

  speed_up() {
    if (this.speed < 8) {
      this.speed += 0.25;
      this.simulate_interval = 500 / this.speed;
      document.getElementById("speed").innerHTML = this.speed.toFixed(2);
    }
  }

  slow_down() {
    if (this.speed > 0.3) {
      this.speed -= 0.25;
      this.simulate_interval = 500 / this.speed;
      document.getElementById("speed").innerHTML = this.speed.toFixed(2);
    }
  }

  drawStateTransPic() {
    var viz = new Viz();
    viz.renderImageElement(this.getContentFromRule())
      .then(function (element) {
        var div = document.getElementById("profile_center");
        div.innerHTML = "";
        div.appendChild(element);
      })
      .catch(error => {
        viz = new Viz();
        console.error(error);
      });
  }
  getContentFromRule() {
    if (!this.is_initialized) {
      alert("Not initialized, please click Set");
      return;
    }
    var content = "digraph turing1 {\n\trankdir = LR;\n";
    var rule_dic = new Array();
    for (var rule of this.setting.rules.data) {
      console.log(rule);
      var lable = rule.data_in + ", " + rule.data_out + ", " + rule.operation;
      var arrow = rule.cur_state + " -> " + rule.next_state;
      if (rule_dic.hasOwnProperty(arrow)) {
        rule_dic[arrow].push(lable);
      } else {
        rule_dic[arrow] = [lable];
      }
    }

    for (let key in rule_dic) {
      var line = key + '[label="';
      for (let lable_index in rule_dic[key]) {
        console.log("lable_index", lable_index);
        if (Number(lable_index) == (rule_dic[key].length - 1)) {
          line = line + " " + rule_dic[key][lable_index] + '"]\n';
        } else {
          line = line + " " + rule_dic[key][lable_index] + "\n";
        }
      }
      content = content + line;
    }
    content = content + "}";

    console.log(content);
    return content;
  }

  async download_rule(need_confirm) {
    if (need_confirm) {
      if (!confirm("Confirm download? Currently edited content will be overwrite")) {
        return;
      }
    }
    // TODO add download logic
    alert("TODO add download logic")
 

  }

  async upload_rule() {
    if (!confirm("Confirm upload rules？")) {
      return;
    }
    // TODO add download logic
    alert("TODO add upload logic")
  }
  // Store parameter values on URL changes
  getRouteParams() {
    console.log("getRouteParams")
    // Route parameters
    this.activatedRoute.params.subscribe(params => {
      this.routeParams = params;
    });

    // URL query parameters
    this.activatedRoute.queryParams.subscribe(params => {
      this.queryParams = params;
    });
  }

  test(type) {
    console.log(type);
    var rules = get_Rules(this.test_raw_rules);
    if (rules == null) {
      return;
    }
    this.init_set();
    var test_data;
    if (type == "1") {
      test_data = "11+111=";
    } else if (type == "2") {
      test_data = "1000+1000=";
    } else if (type == "3") {
      test_data = "1000+1=";
    } else {
      console.log("type error:", type);
      return;
    }

    var tape;
    tape = new Turing_Tape(test_data);

    var step, current_state, data_in;
    step = 0;
    try {
      current_state = "q0";
      data_in = tape.currentCharacter();
   

      while (true) {
        if (current_state == "qa" || step > 10000) {
          if (step > 10000) {
            throw "he number of steps exceeds 10,000 times, it may not be possible to stop";
          }
          break;
        }

        step++;
        
        var current_rule = rules.find_rule(current_state, data_in);
        if (current_rule == null) {
          throw "No corresponding rule found" +
          "State: " +
          current_state +
          " Symbol read: " +
          data_in;
        }
        tape.write(current_rule.data_out);
        tape.move(current_rule.operation);
        current_state = current_rule.next_state;
        data_in = tape.currentCharacter();
      }
      print_current_step(step);
      var answer = "";
      for (let i = 0; i <= 50; i++) {
        answer += tape.characterAtIndex(i);
      }

      var correct = false;
      if (type == "1") {
        if (answer.slice(7, 12) == '11111' && answer[12] == 'B')
          correct = true;
      } else if (type == "2") {
        if (answer.slice(10, 15) == '10000' && answer[15] == 'B')
          correct = true;
      } else if (type == "3") {
        if (answer.slice(7, 11) == '1001' && answer[11] == 'B')
          correct = true;
      } else {
        console.log("type error:", type);
        return;
      }
      var info;
      if (correct) {
        info = "Correct, pass the test sample, the result is: " + answer;
      } else {
        info = "Error, failed the test sample, the result is: " + answer;
      }

      alert(info);
    } catch (error) {
      alert(error);
      return;
    }
  }

}

class Rule {
  cur_state;
  data_in;
  data_out;
  operation;
  next_state;
  constructor(cur_state, data_in, data_out, operation, next_state) {
    this.cur_state = cur_state;
    this.data_in = data_in;
    this.data_out = data_out;
    this.operation = operation;
    this.next_state = next_state;
  }
  print = function () {
    console.log(
      this.cur_state,
      this.data_in,
      this.data_out,
      this.operation,
      this.next_state
    );
  };
  return_str = function () {
    return (
      this.cur_state +
      ", " +
      this.data_in +
      ", " +
      this.data_out +
      ", " +
      this.operation +
      ", " +
      this.next_state
    );
  };
}

class Rules {
  name = "Rules";
  data = [];
  reset = function () {
    return (this.data = []);
  };

  add_rule = function (rule) {
    if (this.find_rule(rule.cur_state, rule.data_in)) {
      alert("same rule: " + rule.return_str());
      return -1;
    } else {
      // console.log("add rule", typeof rule);
      return this.data.push(rule);
    }
  };

  find_rule = function (cur_state, data_in) {
    for (let rule in this.data) {
      if (
        this.data[rule].cur_state == cur_state &&
        this.data[rule].data_in == data_in
      ) {
        return this.data[rule];
      }
    }

    return null;
  };

  print = function () {
    for (let rule in this.data) {
      // console.log(this.data[rule]);
      this.data[rule].print();
    }
  };
}

class set_env {
  rules;
  test_data;
  tape;
  is_pause;
  step;
  current_state;
  constructor(rules) {
    this.rules = get_Rules(rules);
    this.test_data = (document.getElementById("test_data") as HTMLInputElement).value;
    this.tape = new Turing_Tape(this.test_data);
    this.is_pause = true;
    this.step = 0;
    this.current_state = "q0";
    draw_Tape(this.tape);
  }

}

class Turing_Tape {
  currentPos: number;
  borderPos: number;
  printedCharacters;
  constructor(initialdata) {
    this.currentPos = 0;
    this.borderPos = 0;
    var tmp = initialdata.split('');
    this.printedCharacters = [];
    var i: number;
    for (i = -10100; i <= 10100; i++) this.printedCharacters[i] = 'B'
    for (i = 0; i < tmp.length; i++) this.printedCharacters[i] = tmp[i];
  }

  move = function (direction) {
    switch (direction) {
      case "N":
        break;
      case "L":
        this.currentPos--;
        if (this.currentPos < this.borderPos) this.borderPos--;
        break;
      case "R":
        this.currentPos++;
        if (this.currentPos >= this.borderPos + 20) this.borderPos++;
        break;
      default:
        throw "illegal direction input";
    }
  }

  write = function (ch) {
    this.printedCharacters[this.currentPos] = ch;
  }

  currentCharacter = function () {
    return this.printedCharacters[this.currentPos];
  };

  characterAtIndex = function (index) {
    if (this.printedCharacters[index] != null) {
      return this.printedCharacters[index];
    }
    else {
      return "B";
    }
  };

  get_currentPos = function () {
    return this.currentPos - this.borderPos;
  }


}


function get_Rules(test_raw_rules) {
  var rules = new Rules();
  try {
    // var test_raw_rules = rules
    console.log(test_raw_rules)
    var test_rules = test_raw_rules.split("\n");
    // console.log(test_rules);
    for (let index in test_rules) {
      var test_rule = test_rules[index];
      var test_rule_list = test_rule.split(",");
      // console.log(test_rule_list);
      if (test_rule_list.length == 1) {
        if (test_rule_list[0].trim() == "")
          throw "The rule format is incorrect:  empty rule"
      }
      if (test_rule_list.length != 5) {
        throw "The rule format is incorrect: " + test_rule + test_rule_list.length;
      }
      for (let i = 0; i < 5; i++) {
        test_rule_list[i] = test_rule_list[i].trim();
        var element = test_rule_list[i].split(" ");

        if (element.length > 1) {
          throw "The format is incorrect, the number of a certain item is greater than 1: " + test_rule;
        }

        test_rule_list[i] = element[0];
        if (test_rule_list[i].length == 0) {
          throw "Incorrect format, one item is missing: " + test_rule;
        }
       
      }
      if (test_rule_list[1].length != 1) { //delete this sentence to support easy input
        
        throw "symbol read must be a single character" + test_rule;
      }
      var data_in = test_rule_list[1].split("/");
      for (let i in data_in) {
        if (
          data_in[i].length != 1 ||
          test_rule_list[2].length != 1 ||
          test_rule_list[3].length != 1
        ) {
          throw "Please check the symbol read, symbol to write, and move format: " + test_rule;
        }
        if (
          test_rule_list[3] != "L" &&
          test_rule_list[3] != "R"
        ) {
          throw 'Move must be one of "L", "R": ' + test_rule;
        }
        var rule = new Rule(
          test_rule_list[0],
          data_in[i],
          test_rule_list[2],
          test_rule_list[3],
          test_rule_list[4]
        );
        rules.add_rule(rule);
      }
    }
  } catch (error) {
    alert(error);
    return null;
  }
  return rules;
}

function draw_TapeRow(y) {
  var context;
  context = (document.getElementById("paperTapeCanvas") as HTMLCanvasElement).getContext("2d");
  context.lineWidth = 3;
  context.strokeStyle = "black";
  context.beginPath();
  context.moveTo(0, y);
  context.lineTo(1002, y);
  context.closePath();
  return context.stroke();
}

function draw_TapeLine(xCoord, drawThick) {
  var context;
  context = (document.getElementById("paperTapeCanvas") as HTMLCanvasElement).getContext("2d");
  context.lineWidth = drawThick ? 3 : 1;
  context.strokeStyle = drawThick ? "red" : "black";
  context.beginPath();
  context.moveTo(xCoord, 0);
  context.lineTo(xCoord, 50);
  context.closePath();
  return context.stroke();
}

function draw_Tape(tape) {
  var context, i, _i, kThickLine;
  kThickLine = true;
  context = (document.getElementById("paperTapeCanvas") as HTMLCanvasElement).getContext("2d");
  context.clearRect(0, 0, context.canvas.width, context.canvas.height);
  context.font = "bold 48px sans-serif";
  for (i = 0; i <= 19; i++) {
    draw_TapeLine(i * 50 + 1, !kThickLine);
    context.fillText(
      tape.characterAtIndex(i + tape.borderPos),
      i * 50 + 11,
      44
    );
  }
  draw_TapeLine(20 * 50 + 1, !kThickLine);

  context.font = "15px Arial sans-serif";
  for (i = 0; i <= 19; i++) {
    var len = getlen(i + tape.borderPos);
    context.fillText(i + tape.borderPos, i * 50 + 28 - len * 5, 69);
  }

  draw_TapeRow(0);
  draw_TapeRow(50);
  draw_TapeLine(tape.get_currentPos() * 50 + 1, kThickLine);
  draw_TapeLine((tape.get_currentPos() + 1) * 50 + 1, kThickLine);
  move_arrow(tape.get_currentPos());
  return;
}

function getlen(x) {
  if (x == 0) return 1;
  var c = 0;
  var y = 1;
  if (x < 0) { x = -x; c++; }
  while (y <= x) {
    c++;
    y *= 10;
  }
  return c;
}

function move_arrow(step) {
  var img = document.getElementById("arrow");
  img.style.left = step * 50 + "px"; //操作img的style属性使之运动

}

function print_current_step(step) {
  document.getElementById("current_step").innerHTML = step;
}

function print_current_rule(rule_list) {
  var rule = rule_list.return_str();
  document.getElementById("current_rule").innerHTML = rule;
}

function print_next_rule(rule_list) {
  if (rule_list == null) {
    document.getElementById("next_rule").innerHTML = "";
  } else {
    var rule = rule_list.return_str();
    document.getElementById("next_rule").innerHTML = rule;
  }
}