import os
import csv
class Rule:
    def __init__(self, cur_state, data_in, data_out, operation, next_state):
        self.cur_state = cur_state
        self.data_in = data_in
        self.data_out = data_out
        self.operation = operation
        self.next_state = next_state
    def print_rule(self):
        print (self.cur_state, self.data_in, self.data_out, self.operation, self.next_state)
    def get_list(self):
        return [self.cur_state, self.data_in, self.data_out, self.operation, self.next_state]
class Rules:
    def __init__(self):
        self.rules = []
        self.state = {}
        self.state_num = 0
        self.rule_err = False
    def print_rules(self):
        for rule in self.rules:
            rule.print_rule()
    def add_rule(self, rule):
        assert(isinstance(rule, Rule))
        if self.find_rule(rule.cur_state, rule.data_in):
            return -1
        self.rules.append(rule)
        if not rule.cur_state in self.state:
            self.state[rule.cur_state] = 1
            self.state_num += 1
        if not rule.next_state in self.state:
            self.state[rule.next_state] = 1
            self.state_num += 1
        return  0
    def find_rule(self, cur_state, data_in):
        for rule in self.rules:
            if cur_state == rule.cur_state and data_in == rule.data_in:
                return rule
        return None
    def get_rules(self, path):
        f = open(path)
        r = f.readline()
        while r:
            if len(r) <= 0 or r[0]=="\r" or r[0] == "\n" or r[0] == "\t":
                print ("continue")
                r = f.readline()
                continue

            r = r.replace(' ','')
            r = r.replace("\t",'')
            try:
                r = r.split()[0].split(',')
            except:
                print (r,len(r))
                print (ord(r[0]))
                exit("error")

            if len(r) != 5 or len(r[1]) != 1 or len(r[2]) != 1 or len(r[3]) != 1:
                self.rule_err = True
                break
            
            if r[3] != 'N' and r[3] != 'R' and r[3]!='L':
                self.rule_err = True
                break
            rule = Rule(r[0], r[1], r[2], r[3], r[4])
            self.add_rule(rule)
            r = f.readline()
        f.close()

class Test_data:
    def __init__(self, question, answer):
        self.question = question
        self.answer = answer

class Evaluate:
    def __init__(self, rule_path, test_path, maxStep=20000):
        self.rules = Rules()
        self.rules.get_rules(rule_path)
        self.test_data = []
        self.get_test_data(test_path)
        self.test_num = 0
        self.test_step = []
        self.check = {}
        self.reorder_rules = []
        self.maxStep = maxStep

    def get_test_data(self, test_data_path):
        f = open(test_data_path)
        test_data = f.readline().split()
        #test_data should be something like ['$1010111+11101=', 'answer']
        self.test_case_num = 0
        while test_data:
            self.test_case_num += 1
            if len(test_data) == 2:
                data = Test_data(test_data[0], test_data[1])
            else:
                data = Test_data(test_data[0], "")
            self.test_data.append(data)
            test_data = f.readline().split()
    
    def evaluate(self):
        total_score = 0
        record = {
                "detail":{},
                "all":{}}
        for data in self.test_data:
            result = self.test(data.question, data.answer)
            record["detail"][str(self.test_num+1)]={
                "reason":result[1],
                "testcase":result[2]
            } 
            if result[0]: 
                total_score += 1
                record["detail"][str(self.test_num+1)]["state"] = "passed"
            else:
                record["detail"][str(self.test_num+1)]["state"] = "failed"

            self.test_num += 1
        
        import math
        record["all"] = {
            "score": math.ceil(total_score / self.test_num * 100) ,
            "stateNum":self.rules.state_num,
            "totalStep":sum(self.test_step),
            "reason":""
        }
        return record
    def test(self, test_data, correct_answer):
        pointer = self.maxStep
        test_data_record = test_data
        
        test_B = ['B' for i in range(self.maxStep)]
        test_start_point = self.maxStep+len(test_data)
        test_data = test_B + list(test_data) + test_B

        current_state = 'q0'
        step = 0
        if self.rules.rule_err:
            return [False, "rule error",test_data_record]
        while not current_state == 'qa' and not step >= self.maxStep:
            step += 1
            rule = self.rules.find_rule(current_state, test_data[pointer])

            if not rule:
                return [False, "can't find rule at state: "+current_state+",input: "+test_data[pointer], test_data_record]
            if (not self.check.__contains__(str(rule.get_list()))):
                self.check[str(rule.get_list())] =True
                self.reorder_rules.append(",".join(rule.get_list()))

            test_data[pointer] = rule.data_out
            if rule.operation == 'L':
                pointer -= 1
            elif rule.operation == 'R':
                pointer += 1

            if pointer < 0:
                return [False, "pointer can't be nagative", test_data_record]
 
            current_state = rule.next_state
        
        self.test_step.append(step)
        if step > self.maxStep:
            return [0, "step excced "+str(self.maxStep),test_data_record]


        stu_ans = ''

        for i in range(test_start_point, len(test_data), 1):

            if test_data[i] != 'B':
                stu_ans += test_data[i]
            else:
                break
  
        if stu_ans == correct_answer:
            return [True, "correct",test_data_record]
        else:
            return [False, "wrong, student answer: "+stu_ans+"correct answer: "+correct_answer,  test_data_record]

if __name__ =="__main__":
    def exp_eval(exp_path):
        parents = os.listdir(exp_path)
        # print parents
        students_score = []
        first_check = True
        for parent in parents:
            if len(parent) < 4 or parent[-4:] != '.txt':
                continue
            name = parent[:-4]
            child = os.path.join(exp_path,parent)   
            test_path = os.path.join(exp_path, 'test_data')
            eval = Evaluate(child, test_path)
            if first_check:
                first_check = False
                test_case_num = eval.test_case_num
                first_row = [ "name"]
                for i in range(test_case_num):
                    first_row += ["test case "+str(i+1), "result " + str(i+1)]
                first_row+=["total_score", "state num", "total_step"]
                students_score.append(first_row)
            eval_result = eval.evaluate()
            result = [name]
            for i in range(test_case_num):
                result.append(eval_result["detail"][str(i+1)]["testcase"])
                      # result += [eval[]]      result.append(eval_result["detail"][str(i+1)]["reason"])
            result.append(eval_result["all"]["score"])
            result.append(eval_result["all"]["stateNum"])
            result.append(eval_result["all"]["totalStep"])
            students_score.append(result)
            
        out = open(os.path.join(exp_path,exp_path+'.csv'),'w')
        csv_write = csv.writer(out,dialect='excel')
        for score in students_score:
            csv_write.writerow(score)
        out.close()
    
    childs = os.listdir("./")
    for child in childs:
        if os.path.isdir(child):
            print ("test "+child)
            try:
                exp_eval(child)
            except:
                pass
    